// main.c


// Contains the .dll entry point and overridden functions
// uses imports.c for loading symbols, dib.c for getting framebuffer access.

// Copr. 1998 - Justin Frankel/Nullsoft, Inc. (http://www.nullsoft.com)

// Version 0.5 - 1/4/1998

/* changes by Gerhard Pfeiffer ( gpfeiffer@carinthia.com )
   *  now the wglide library tries to emulate the voodoo rush chipset in
      window mode.
      ( it overwrites the chiptype )
   *  window handling ( resizing ) is implemented ( should work with mesa )
   *  a 800x600 windowmode including z-Buffer is implemented,
      but hangs the system within few seconds.
*/

// The highreshack enables max resolution 800x600 on a Voodoo with only 
// 2 MByte framebuffer by disabling the frontbuffer
// doesnt work correctly, most applications hang within a few seconds
// maybe i have done something wrong here ?

// #define highreshack

#include <windows.h>

#define FX_GLIDE_NO_FUNC_PROTO
#include <glide.h>
#include <glideutl.h>

#include "pointers.h"

HINSTANCE hGlide, hDllInst;

#define ENTRY __declspec( dllexport )
#define CALL __stdcall

// dib.c functions
extern void dibinit(HWND hwnd, int w, int h);
extern void dibquit();
extern short *dibgetbuf();
extern void dibupdate();
extern int dibgetpitch();

// imports.c functions
extern int LoadSymbols(HINSTANCE *hGlide, HINSTANCE hDllInst);

// DLL entry point. Calls on load/unload of the .dll. Loads or frees the real glide2x.dll.
BOOL WINAPI dllmain(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	hDllInst = hInst;
	if (ul_reason_for_call == DLL_PROCESS_ATTACH)
	{
		if (!LoadSymbols(&hGlide,hDllInst)) 
		{
			hGlide = NULL;
			return FALSE;
		}
	}
	if (ul_reason_for_call == DLL_PROCESS_DETACH)
	{
		FreeModule(hGlide);
		hGlide = NULL;
	}
	return TRUE;
}

// hwnd is no defined global, for later use in grSSTControl
HWND hwnd; 
GrOriginLocation_t wmesa_origin_location;

int enable=1, // are we running windowed?
    sw, // screen width (window width)
	sh, // screen height (window height)
	offset_forlowresmodesinquake2; // quake 2 320x240 and 400x300 hack.
	glideyres,glidexres;


// Our new, overridden functions go here. 


// override grSstWinOpen, for opening the 3dfx hardware.
ENTRY FxBool CALL grSstWinOpen(
          FxU32                hWnd,
          GrScreenResolution_t screen_resolution,
          GrScreenRefresh_t    refresh_rate,
          GrColorFormat_t      color_format,
          GrOriginLocation_t   origin_location,
          int                  nColBuffers,
          int                  nAuxBuffers)
{
	hwnd=(HWND)hWnd; // get hWnd as a HWND.
	offset_forlowresmodesinquake2=0; // default: set it to 0 for 
	// save origionlocation for later use ( in grSstControl )
	wmesa_origin_location=origin_location;
	enable=0;
	// decide if we are running fullscreen or windowed
    if (hwnd && IsWindow(hwnd)) 
	{	// referring to the glide manual, windowed rendering can be activated
		// by setting screen_resolution to GR_RESOLUTION_NONE 
		if(screen_resolution==GR_RESOLUTION_NONE) 
		{	RECT r;
			enable = 1;
#ifdef highreshack 
			// highreshack, enables max Resolution 800x600,
			// only uses one framebuffer ( frontbuffer )
			screen_resolution=GR_RESOLUTION_800x600;
			glideyres=600;glidexres=800;
			nColBuffers=1;  
#else 
			// always use GR_RESOLUTION_640x480 in window mode
			screen_resolution=GR_RESOLUTION_640x480;
			glideyres=480;glidexres=640;
#endif
			// calculate the correct offset
			GetClientRect(hwnd,&r);
	        if (wmesa_origin_location == GR_ORIGIN_LOWER_LEFT) 
				
			sw = r.right;if(sw>glidexres) sw=glidexres;
		    sh = r.bottom;if(sh>glideyres) sh=glideyres;
			if (wmesa_origin_location == GR_ORIGIN_LOWER_LEFT) 
				offset_forlowresmodesinquake2=glideyres-sh;

		}
	} 

    // initalize voodoo
	if (_grSstWinOpen(hWnd,screen_resolution,refresh_rate,color_format,origin_location,nColBuffers,nAuxBuffers) == FXFALSE)

    return FXFALSE;
	
	// default height and width ( not true if window rendering is enabled )
	if(!enable)
	{ sw=_grSstScreenWidth();
	  sh=_grSstScreenHeight();
	} else
	{	_grSstControl(GR_CONTROL_DEACTIVATE);
#ifdef highreshack
		// _grRenderBuffer(GR_BUFFER_FRONTBUFFER);
		_grRenderBuffer(GR_BUFFER_BACKBUFFER);
#else
		_grRenderBuffer(GR_BUFFER_BACKBUFFER);
#endif
		dibinit(hwnd,glidexres,glideyres);
	}
	return FXTRUE;
}

// override grBufferSwap to copy the hardware's data into our DIBSection
ENTRY void CALL grBufferSwap( int swap_interval )
{
	if (enable)
	{
		short *data = dibgetbuf();
		if (data)
		{
#ifdef highreshack 			
			_grLfbReadRegion(GR_BUFFER_FRONTBUFFER,0,offset_forlowresmodesinquake2,sw,sh,dibgetpitch(),data);
			_grLfbReadRegion(GR_BUFFER_BACKBUFFER,0,offset_forlowresmodesinquake2,sw,sh,dibgetpitch(),data);
			dibupdate();
			_grRenderBuffer(GR_BUFFER_BACKBUFFER);
#else
			_grBufferSwap(swap_interval );
			_grLfbReadRegion(GR_BUFFER_FRONTBUFFER,0,offset_forlowresmodesinquake2,sw,sh,dibgetpitch(),data);
			dibupdate();
		    // _grLfbReadRegion(GR_BUFFER_BACKBUFFER,0,offset_forlowresmodesinquake2,sw,sh,dibgetpitch(),data);
#endif
		}
	}
	else _grBufferSwap(swap_interval);
}

// override grSstWinClose to delete our DIB
ENTRY void CALL grSstWinClose( void )
{
	_grSstWinClose();
	if (enable) dibquit();
}

// override grSstControl to keep the calling app from reenabling the hardware's video output
// and implement the window handling code
ENTRY FxBool CALL grSstControl( FxU32 code )
{
	if (!enable) return _grSstControl(code);
	if (code == GR_CONTROL_RESIZE)
	{ // Window resizing
		RECT r;
		GetClientRect(hwnd,&r);
		sw = r.right; if(sw>glidexres) sw=glidexres;
		sh = r.bottom;if(sh>glideyres) sh=glideyres;
		if (wmesa_origin_location == GR_ORIGIN_LOWER_LEFT) 
			offset_forlowresmodesinquake2=glideyres-sh;
		return 1;
	}
	if (code == GR_CONTROL_MOVE) return 1;
	if (code == GR_CONTROL_DEACTIVATE) return 1;
    if (code == GR_CONTROL_ACTIVATE) return 1;
	return 0;
}

// override grSstControl to return a type which can render into a window
// ( voodoo rush )
ENTRY FxBool CALL grSstQueryHardware( GrHwConfiguration *hwconfig )
{ 
	FxBool ret;
	ret=_grSstQueryHardware( hwconfig );
	if(ret) hwconfig->SSTs[0].type=GR_SSTTYPE_SST96;
	return ret;
}

// The new minigl port seems to use this function different with the
// voodoo rush chipset, so it doesnt work
// use an older miniport here
ENTRY void grTexDownloadMipMap( GrChipID_t tmu,FxU32  startAddress,FxU32 evenOdd,GrTexInfo *info )
{ // if(startAddress<2090000)
  _grTexDownloadMipMap( tmu, startAddress, evenOdd, info);
}

ENTRY void grRenderBuffer( GrBuffer_t buffer )
{  
#ifdef highreshack 
  // do nothing
#else
      _grRenderBuffer( buffer );
#endif
}